@@ -351,12 +351,16 @@ def cidr_make_var(val, parent_cidr_var, exp, lang, register=True, raw=False):
     pretty_printer, dynamic = get_printer(val, raw)
     # non-dynamic reference vars show referenced value
     # see c_value_of_variable()
+    value_pretty_printer = pretty_printer
     if not dynamic and val_type_no_typedefs.code in (gdb.TYPE_CODE_REF, gdb.TYPE_CODE_RVALUE_REF):
         print_val_type = gdb.types.get_basic_type(val_type_no_typedefs)
         print_val = val.cast(print_val_type)
         value_pretty_printer, dynamic = get_printer(print_val, raw)
-    else:
-        value_pretty_printer = pretty_printer
+    elif not dynamic and val_type.code == gdb.TYPE_CODE_TYPEDEF:
+        no_typedef_printer, no_typedef_printer_found = get_printer(val.cast(val_type_no_typedefs), raw)
+        if no_typedef_printer_found:
+            value_pretty_printer = no_typedef_printer
+            dynamic = True
 
     # When printing children of a pointer to a struct/union we want to
     # remove one layer of indirection and show fields of the struct/union.
@@ -365,21 +369,12 @@ def cidr_make_var(val, parent_cidr_var, exp, lang, register=True, raw=False):
     children_val_basic_type = gdb.types.get_basic_type(val_type) # strips typedefs and refs
     children_val = val.cast(children_val_basic_type)
     children_val_pretty_printer, children_val_dynamic = get_printer(children_val, raw)
-    if children_val_basic_type.code == gdb.TYPE_CODE_PTR:
+    if children_val_basic_type.code == gdb.TYPE_CODE_PTR and children_val != 0:
         target_type = gdb.types.get_basic_type(children_val_basic_type.target())
         if target_type.code in (gdb.TYPE_CODE_STRUCT, gdb.TYPE_CODE_UNION):
-            try:
-                referenced_value = children_val.referenced_value()
-                # Listing children of a null pointer to a struct/union will fail with memory error in _some_ cases.
-                # Fetch the referenced value now to ensure the pointer is not null.
-                # If the pointer is null, we will not update children_val/children_val_pretty_printer
-                # and will not list struct/union fields.
-                referenced_value.fetch_lazy()
-                children_val = referenced_value
-                children_val_pretty_printer, children_val_dynamic = get_printer(children_val, raw)
-            except gdb.MemoryError:
-                # pointer dereference failed, will not list struct/union fields
-                pass
+            referenced_value = children_val.referenced_value()
+            children_val = referenced_value
+            children_val_pretty_printer, children_val_dynamic = get_printer(children_val, raw)
 
     dynamic |= children_val_dynamic
 
@@ -406,6 +401,11 @@ def cidr_make_var(val, parent_cidr_var, exp, lang, register=True, raw=False):
     value = None
     if hasattr(value_pretty_printer, 'to_string'):
         value = CidrVarCreate._to_string_runner(value_pretty_printer.to_string)
+        if val_type_no_typedefs.code == gdb.TYPE_CODE_PTR and not raw and isinstance(value_pretty_printer, default.printers.PointerPrinter):
+            ptr_target_printer, ptr_target_dynamic = get_printer(val.referenced_value(), raw)
+            if ptr_target_dynamic and hasattr(ptr_target_printer, 'to_string'):
+                ptr_target_value = CidrVarCreate._to_string_runner(ptr_target_printer.to_string)
+                value = "%s %s" % (value, ptr_target_value)
     elif hasattr(children_val_pretty_printer, 'children'):
         # see varobj_value_get_print_value()
         value = "{...}"
@@ -591,6 +591,70 @@ def make_nop_visualizer(value):
         result = gdb.printing.NoOpScalarPrinter(value)
     return result
 
+# _cidr_finish_breakpoints contains temporary breakpoints created by
+# -cidr-exec-step and -cidr-exec-next commands.
+# We need to track and delete breakpoints manually because they are
+# not removed automatically if not reached.
+# Cleanup happens every time the program stops.
+_cidr_finish_breakpoints = []
+
+def _clear_finish_breakpoints(event):
+    for bp in _cidr_finish_breakpoints:
+        if bp.is_valid():
+            bp.delete()
+    _cidr_finish_breakpoints.clear()
+
+
+def _find_frame_for_finish_breakpoint():
+    """
+    Returns a frame for which the caller frame has sources
+    or None if no such frame found.
+    """
+    frame = gdb.newest_frame()
+    while frame is not None and frame.is_valid():
+        callerFrame = frame.older()
+        if callerFrame is not None and callerFrame.is_valid():
+            sal = callerFrame.find_sal()
+            if sal.is_valid() and sal.symtab is not None:
+                return frame
+        frame = callerFrame
+    return None
+
+
+class CidrExecNextCommand(gdb.MICommand):
+    """
+    CidrExecNextCommand is similar to the -exec-next gdb/mi command,
+    but also stops on exit from a function like -exec-finish does.
+    """
+
+    def __init__(self):
+        super(CidrExecNextCommand, self).__init__("-cidr-exec-next")
+
+    def invoke(self, argv):
+        frame = _find_frame_for_finish_breakpoint()
+        if frame is not None:
+            _cidr_finish_breakpoints.append(gdb.FinishBreakpoint(frame, internal=True))
+        gdb.execute_mi("-exec-next")
+        return None
+
+
+class CidrExecStepCommand(gdb.MICommand):
+    """
+    CidrExecStepCommand is similar to the -exec-step gdb/mi command,
+    but also stops on exit from a function like -exec-finish does.
+    """
+
+    def __init__(self):
+        super(CidrExecStepCommand, self).__init__("-cidr-exec-step")
+
+    def invoke(self, argv):
+        frame = _find_frame_for_finish_breakpoint()
+        if frame is not None:
+            _cidr_finish_breakpoints.append(gdb.FinishBreakpoint(frame, internal=True))
+        gdb.execute_mi("-exec-step")
+        return None
+
+
 # note that register_commands will fail for old gdb versions not supporting gdb.MICommand
 # it is responsibility of the caller to check the version
 def register_commands():
@@ -609,3 +673,7 @@ def register_commands():
     CidrVarDeleteAll()
     CidrVarRead()
     CidrVarSetRaw()
+
+    CidrExecNextCommand()
+    CidrExecStepCommand()
+    gdb.events.stop.connect(_clear_finish_breakpoints)
